@@ -11,6 +11,8 @@
 #ifndef __G_PASTE_GDBUS_MACROS_H__
 #define __G_PASTE_GDBUS_MACROS_H__
 
+#ifndef __GI_SCANNER__
+
 #include <gpaste-gdbus-defines.h>
 #include <gpaste-util.h>
 
@@ -33,36 +35,57 @@ G_BEGIN_DECLS
                                 "g-interface-name", G_PASTE_##BUS_ID##_INTERFACE_NAME, \
                                 NULL)
 
+#define CUSTOM_PROXY_RET(TYPE) \
+    if (_error)                \
+    {                          \
+        if (error)             \
+        {                      \
+            *error = _error;   \
+            _error = NULL;     \
+        }                      \
+        return NULL;           \
+    }                          \
+    return (self) ? G_PASTE_##TYPE (self) : NULL
+
 #define CUSTOM_PROXY_NEW_FINISH(TYPE)                                       \
     g_return_val_if_fail (G_IS_ASYNC_RESULT (result), NULL);                \
     g_return_val_if_fail (!error || !(*error), NULL);                       \
     g_autoptr (GObject) source = g_async_result_get_source_object (result); \
+    g_autoptr (GError) _error = NULL;                                       \
     g_assert (source);                                                      \
     GObject *self = g_async_initable_new_finish (G_ASYNC_INITABLE (source), \
                                                  result,                    \
-                                                 error);                    \
-    return (self) ? G_PASTE_##TYPE (self) : NULL;
+                                                 &_error);                  \
+    CUSTOM_PROXY_RET (TYPE);
 
 #define CUSTOM_PROXY_NEW(TYPE, BUS_ID, BUS_NAME)                                             \
+    g_autoptr (GError) _error = NULL;                                                        \
     GInitable *self = g_initable_new (G_PASTE_TYPE_##TYPE,                                   \
                                       NULL, /* cancellable */                                \
-                                      error,                                                 \
+                                      &_error,                                               \
                                       "g-bus-type",       G_BUS_TYPE_SESSION,                \
                                       "g-flags",          G_DBUS_PROXY_FLAGS_NONE,           \
                                       "g-name",           BUS_NAME,                          \
                                       "g-object-path",    G_PASTE_##BUS_ID##_OBJECT_PATH,    \
                                       "g-interface-name", G_PASTE_##BUS_ID##_INTERFACE_NAME, \
                                       NULL);                                                 \
-    return (self) ? G_PASTE_##TYPE (self) : NULL;
+    CUSTOM_PROXY_RET (TYPE);
 
 /********************/
 /* Methods / Common */
 /********************/
 
-#define DBUS_PREPARE_EXTRACTION                      \
-        GVariantIter result_iter;                    \
-        g_variant_iter_init (&result_iter, _result); \
-        g_autoptr (GVariant) variant = g_variant_iter_next_value (&result_iter)
+#define DBUS_PREPARE_EXTRACTION(iter)                                  \
+    g_autoptr (GVariant) _variant = NULL;                              \
+    G_GNUC_UNUSED GVariant *variant;                                   \
+    if (iter)                                                          \
+    {                                                                  \
+        GVariantIter result_iter;                                      \
+        g_variant_iter_init (&result_iter, _result);                   \
+        variant = _variant = g_variant_iter_next_value (&result_iter); \
+    }                                                                  \
+    else                                                               \
+        variant = _result
 
 #define DBUS_RETURN(if_fail, extract_and_return_answer) \
     if (!_result)                                       \
@@ -74,7 +97,7 @@ G_BEGIN_DECLS
 /*****************************/
 
 #define DBUS_CALL_ASYNC_FULL(TYPE_CHECKER, decl, method, params, n_params) \
-    g_return_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self));                   \
+    g_return_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self));                  \
     decl;                                                                  \
     g_dbus_proxy_call (G_DBUS_PROXY(self),                                 \
                        method,                                             \
@@ -115,14 +138,17 @@ G_BEGIN_DECLS
                                                              error);              \
     DBUS_RETURN (if_fail, extract_and_return_answer)
 
-#define DBUS_ASYNC_FINISH_WITH_RETURN(TYPE_CHECKER, if_fail, extract_and_return_answer)                \
-    DBUS_ASYNC_FINISH_FULL (g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), if_fail);          \
-                            g_return_val_if_fail (G_IS_ASYNC_RESULT (result), if_fail);                \
-                            g_return_val_if_fail (!error || !(*error), if_fail),                       \
-                            if_fail,                                                                   \
-                            DBUS_PREPARE_EXTRACTION;                                                   \
+#define DBUS_ASYNC_FINISH_WITH_RETURN_FULL(TYPE_CHECKER, if_fail, iter, extract_and_return_answer) \
+    DBUS_ASYNC_FINISH_FULL (g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), if_fail);     \
+                            g_return_val_if_fail (G_IS_ASYNC_RESULT (result), if_fail);            \
+                            g_return_val_if_fail (!error || !(*error), if_fail),                   \
+                            if_fail,                                                               \
+                            DBUS_PREPARE_EXTRACTION(iter);                                         \
                             extract_and_return_answer)
 
+#define DBUS_ASYNC_FINISH_WITH_RETURN(TYPE_CHECKER, if_fail, extract_and_return_answer) \
+    DBUS_ASYNC_FINISH_WITH_RETURN_FULL(TYPE_CHECKER, if_fail, TRUE, extract_and_return_answer)
+
 /***********************************/
 /* Methods / Async / Impl - Finish */
 /***********************************/
@@ -144,11 +170,14 @@ G_BEGIN_DECLS
 #define DBUS_ASYNC_FINISH_RET_STRING_BASE(TYPE_CHECKER) \
     DBUS_ASYNC_FINISH_WITH_RETURN (TYPE_CHECKER, NULL, return g_variant_dup_string (variant, NULL))
 
+#define DBUS_ASYNC_FINISH_RET_ITEM_BASE(TYPE_CHECKER) \
+    DBUS_ASYNC_FINISH_WITH_RETURN_FULL (TYPE_CHECKER, NULL, FALSE, return g_paste_util_get_dbus_item_result (variant))
+
 #define DBUS_ASYNC_FINISH_RET_STRV_BASE(TYPE_CHECKER) \
     DBUS_ASYNC_FINISH_WITH_RETURN (TYPE_CHECKER, NULL, return g_variant_dup_strv (variant, NULL))
 
-#define DBUS_ASYNC_FINISH_RET_AT_BASE(TYPE_CHECKER, len) \
-    DBUS_ASYNC_FINISH_WITH_RETURN (TYPE_CHECKER, NULL, return g_paste_util_get_dbus_at_result (variant, len))
+#define DBUS_ASYNC_FINISH_RET_ITEMS_BASE(TYPE_CHECKER) \
+    DBUS_ASYNC_FINISH_WITH_RETURN (TYPE_CHECKER, NULL, return g_paste_util_get_dbus_items_result (variant))
 
 #define DBUS_ASYNC_FINISH_RET_AU_BASE(TYPE_CHECKER, len) \
     DBUS_ASYNC_FINISH_WITH_RETURN (TYPE_CHECKER, NULL, return g_paste_util_get_dbus_au_result (variant, len))
@@ -180,16 +209,19 @@ G_BEGIN_DECLS
 /* Methods / Sync / General - With return */
 /******************************************/
 
-#define DBUS_CALL_WITH_RETURN_FULL_BASE(TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract, pre_extract)                        \
+#define DBUS_CALL_WITH_RETURN_FULL_BASE(TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract, pre_extract)    \
     DBUS_CALL_FULL (g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), if_fail), decl, method, params, n_params, if_fail, \
-                    pre_extract;                                                                                                                    \
+                    pre_extract;                                                                                                \
                     variant_extract)
 
 #define DBUS_CALL_WITH_RETURN_RAW_BASE(TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract) \
     DBUS_CALL_WITH_RETURN_FULL_BASE (TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract, GVariant *variant = _result)
 
+#define DBUS_CALL_WITH_RETURN_BASE_FULL(TYPE_CHECKER, decl, method, params, n_params, if_fail, iter, variant_extract) \
+    DBUS_CALL_WITH_RETURN_FULL_BASE (TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract, DBUS_PREPARE_EXTRACTION(iter))
+
 #define DBUS_CALL_WITH_RETURN_BASE(TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract) \
-    DBUS_CALL_WITH_RETURN_FULL_BASE (TYPE_CHECKER, decl, method, params, n_params, if_fail, variant_extract, DBUS_PREPARE_EXTRACTION)
+    DBUS_CALL_WITH_RETURN_BASE_FULL(TYPE_CHECKER, decl, method, params, n_params, if_fail, TRUE, variant_extract)
 
 /*************************************/
 /* Methods / Sync / Impl - No return */
@@ -218,19 +250,19 @@ G_BEGIN_DECLS
     DBUS_CALL_WITH_RETURN_BASE (TYPE_CHECKER, {}, method, NULL, 0, if_fail, variant_extract)
 
 #define DBUS_CALL_NO_PARAM_RET_STRING_BASE(TYPE_CHECKER, method) \
-    DBUS_CALL_NO_PARAM_BASE(TYPE_CHECKER, method, NULL, return g_variant_dup_string (variant, NULL)) \
+    DBUS_CALL_NO_PARAM_BASE (TYPE_CHECKER, method, NULL, return g_variant_dup_string (variant, NULL))
 
 #define DBUS_CALL_NO_PARAM_RET_STRV_BASE(TYPE_CHECKER, method) \
-    DBUS_CALL_NO_PARAM_BASE(TYPE_CHECKER, method, NULL, return g_variant_dup_strv (variant, NULL)) \
+    DBUS_CALL_NO_PARAM_BASE (TYPE_CHECKER, method, NULL, return g_variant_dup_strv (variant, NULL))
 
-#define DBUS_CALL_ONE_PARAMV_RET_AT_BASE(TYPE_CHECKER, method, paramv, len) \
-    DBUS_CALL_WITH_RETURN_BASE(TYPE_CHECKER, {}, method, &paramv, 1, NULL, return g_paste_util_get_dbus_at_result (variant, len))
+#define DBUS_CALL_NO_PARAM_RET_ITEMS_BASE(TYPE_CHECKER, method) \
+    DBUS_CALL_NO_PARAM_BASE (TYPE_CHECKER, method, NULL, return g_paste_util_get_dbus_items_result (variant))
 
 #define DBUS_CALL_ONE_PARAMV_RET_AU_BASE(TYPE_CHECKER, method, paramv, len) \
-    DBUS_CALL_WITH_RETURN_BASE(TYPE_CHECKER, {}, method, &paramv, 1, NULL, return g_paste_util_get_dbus_au_result (variant, len))
+    DBUS_CALL_WITH_RETURN_BASE (TYPE_CHECKER, {}, method, &paramv, 1, NULL, return g_paste_util_get_dbus_au_result (variant, len))
 
-#define DBUS_CALL_ONE_PARAMV_RET_STRV_BASE(TYPE_CHECKER, method, paramv) \
-    DBUS_CALL_WITH_RETURN_BASE(TYPE_CHECKER, {}, method, &paramv, 1, NULL, return g_variant_dup_strv (variant, NULL))
+#define DBUS_CALL_ONE_PARAMV_RET_ITEMS_BASE(TYPE_CHECKER, method, paramv) \
+    DBUS_CALL_WITH_RETURN_BASE (TYPE_CHECKER, {}, method, &paramv, 1, NULL, return g_paste_util_get_dbus_items_result (variant))
 
 /******************************************************/
 /* Methods / Sync / General - With return - One param */
@@ -239,24 +271,30 @@ G_BEGIN_DECLS
 #define DBUS_CALL_ONE_PARAM_RAW_BASE(TYPE_CHECKER, param_type, param_name, method, if_fail, variant_extract) \
     DBUS_CALL_WITH_RETURN_RAW_BASE (TYPE_CHECKER, GVariant *parameter = g_variant_new_##param_type (param_name), method, &parameter, 1, if_fail, variant_extract)
 
+#define DBUS_CALL_ONE_PARAM_BASE_FULL(TYPE_CHECKER, param_type, param_name, method, if_fail, iter, variant_extract) \
+    DBUS_CALL_WITH_RETURN_BASE_FULL (TYPE_CHECKER, GVariant *parameter = g_variant_new_##param_type (param_name), method, &parameter, 1, if_fail, iter, variant_extract)
+
 #define DBUS_CALL_ONE_PARAM_BASE(TYPE_CHECKER, param_type, param_name, method, if_fail, variant_extract) \
-    DBUS_CALL_WITH_RETURN_BASE (TYPE_CHECKER, GVariant *parameter = g_variant_new_##param_type (param_name), method, &parameter, 1, if_fail, variant_extract)
+    DBUS_CALL_ONE_PARAM_BASE_FULL (TYPE_CHECKER, param_type, param_name, method, if_fail, TRUE, variant_extract)
 
 /***************************************************/
 /* Methods / Sync / Impl - With return - One param */
 /***************************************************/
 
 #define DBUS_CALL_ONE_PARAM_RET_BOOL_BASE(TYPE_CHECKER, param_type, param_name, method) \
-    DBUS_CALL_ONE_PARAM_BASE(TYPE_CHECKER, param_type, param_name, method, FALSE, return g_variant_get_boolean (variant))
+    DBUS_CALL_ONE_PARAM_BASE (TYPE_CHECKER, param_type, param_name, method, FALSE, return g_variant_get_boolean (variant))
 
 #define DBUS_CALL_ONE_PARAM_RET_UINT64_BASE(TYPE_CHECKER, param_type, param_name, method) \
-    DBUS_CALL_ONE_PARAM_BASE(TYPE_CHECKER, param_type, param_name, method, 0, return g_variant_get_uint64 (variant))
+    DBUS_CALL_ONE_PARAM_BASE (TYPE_CHECKER, param_type, param_name, method, 0, return g_variant_get_uint64 (variant))
 
 #define DBUS_CALL_ONE_PARAM_RET_STRING_BASE(TYPE_CHECKER, param_type, param_name, method) \
-    DBUS_CALL_ONE_PARAM_BASE(TYPE_CHECKER, param_type, param_name, method, NULL, return g_variant_dup_string (variant, NULL /* length */))
+    DBUS_CALL_ONE_PARAM_BASE (TYPE_CHECKER, param_type, param_name, method, NULL, return g_variant_dup_string (variant, NULL /* length */))
 
-#define DBUS_CALL_ONE_PARAM_RET_AT_BASE(TYPE_CHECKER, param_type, param_name, method, len) \
-    DBUS_CALL_ONE_PARAM_BASE(TYPE_CHECKER, param_type, param_name, method, NULL, return g_paste_util_get_dbus_at_result (variant, len))
+#define DBUS_CALL_ONE_PARAM_RET_STRV_BASE(TYPE_CHECKER, param_type, param_name, method) \
+    DBUS_CALL_ONE_PARAM_BASE (TYPE_CHECKER, param_type, param_name, method, NULL, return g_variant_dup_strv (variant, NULL))
+
+#define DBUS_CALL_ONE_PARAM_RET_ITEM_BASE(TYPE_CHECKER, param_type, param_name, method) \
+    DBUS_CALL_ONE_PARAM_BASE_FULL (TYPE_CHECKER, param_type, param_name, method, NULL, FALSE, return g_paste_util_get_dbus_item_result (variant))
 
 /****************************************************/
 /* Methods / Sync / Impl - With return - Two params */
@@ -268,15 +306,18 @@ G_BEGIN_DECLS
 #define DBUS_CALL_TWO_PARAMS_RET_UINT64_BASE(TYPE_CHECKER, params, method) \
     DBUS_CALL_TWO_PARAMS_BASE(TYPE_CHECKER, params, method, 0, return g_variant_get_uint64 (variant))
 
-#define DBUS_CALL_TWO_PARAMS_RET_UINT32_BASE(TYPE_CHECKER, params, method) \
-    DBUS_CALL_TWO_PARAMS_BASE(TYPE_CHECKER, params, method, 0, return g_variant_get_uint32 (variant))
+#define DBUS_CALL_THREE_PARAMS_BASE(TYPE_CHECKER, params, method, if_fail, variant_extract) \
+    DBUS_CALL_WITH_RETURN_BASE (TYPE_CHECKER, {}, method, params, 3, if_fail, variant_extract)
+
+#define DBUS_CALL_THREE_PARAMS_RET_UINT32_BASE(TYPE_CHECKER, params, method) \
+    DBUS_CALL_THREE_PARAMS_BASE(TYPE_CHECKER, params, method, 0, return g_variant_get_uint32 (variant))
 
 /************************/
 /* Properties / Getters */
 /************************/
 
 #define DBUS_GET_PROPERTY_INIT(TYPE_CHECKER, property,  _default)                        \
-    g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), _default);                   \
+    g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), _default);                  \
     g_autoptr (GVariant) result = g_dbus_proxy_get_cached_property (G_DBUS_PROXY (self), \
                                                                     property);           \
     if (!result)                                                                         \
@@ -295,7 +336,7 @@ G_BEGIN_DECLS
 /************************/
 
 #define DBUS_SET_GENERIC_PROPERTY_BASE(TYPE_CHECKER, iface, property, value, vtype)              \
-    g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), FALSE);                              \
+    g_return_val_if_fail (_G_PASTE_IS_##TYPE_CHECKER (self), FALSE);                             \
     GVariant *prop[] = {                                                                         \
         g_variant_new_string (iface),                                                            \
         g_variant_new_string (property),                                                         \
@@ -315,4 +356,5 @@ G_BEGIN_DECLS
 
 G_END_DECLS
 
+#endif /*__GI_SCANNER__ */
 #endif /*__G_PASTE_GDBUS_MACROS_H__*/
