@@ -34,7 +34,6 @@ typedef struct
 
     GPasteSettings         *settings;
     GPasteGnomeShellClient *shell_client;
-    gboolean                use_shell_client;
     gboolean                grabbing;
     guint64                 retries;
 
@@ -158,7 +157,6 @@ typedef struct
     GPasteKeybinding       *binding;
     GPasteSettings         *settings;
     GPasteGnomeShellClient *shell_client;
-    gboolean                use_shell_client;
 
     guint32                 action;
 
@@ -199,10 +197,7 @@ _keybinding_grab_gnome_shell (_Keybinding *k)
     if (k->action || !k->shell_client)
         return;
 
-    GPasteGnomeShellAccelerator accel = {
-        g_paste_keybinding_get_accelerator (k->binding, k->settings),
-        G_PASTE_GNOME_SHELL_ACTION_MODE_ALL
-    };
+    GPasteGnomeShellAccelerator accel = G_PASTE_GNOME_SHELL_ACCELERATOR (g_paste_keybinding_get_accelerator (k->binding, k->settings));
 
     g_paste_gnome_shell_client_grab_accelerator (k->shell_client, accel, on_key_grabbed, k);
 }
@@ -211,7 +206,7 @@ static void
 _keybinding_grab (_Keybinding *k)
 {
     _keybinding_activate (k);
-    if (k->use_shell_client)
+    if (k->shell_client)
         _keybinding_grab_gnome_shell (k);
     else
         g_paste_keybinder_change_grab_internal (k->binding, TRUE);
@@ -233,17 +228,17 @@ on_key_ungrabbed (GObject      *source_object,
 static void
 _keybinding_ungrab_gnome_shell (_Keybinding *k)
 {
-    if (k->action)
-    {
-        g_paste_gnome_shell_client_ungrab_accelerator (k->shell_client, k->action, on_key_ungrabbed, NULL);
-        k->action = 0;
-    }
+    if (!k->action || !k->shell_client)
+        return;
+
+    g_paste_gnome_shell_client_ungrab_accelerator (k->shell_client, k->action, on_key_ungrabbed, NULL);
+    k->action = 0;
 }
 
 static void
 _keybinding_ungrab (_Keybinding *k)
 {
-    if (k->use_shell_client)
+    if (k->shell_client)
         _keybinding_ungrab_gnome_shell (k);
     else
         g_paste_keybinder_change_grab_internal (k->binding, FALSE);
@@ -269,7 +264,6 @@ _keybinding_new (GPasteKeybinding       *binding,
     k->binding = binding;
     k->settings = g_object_ref (settings);
     k->shell_client = (shell_client) ? g_object_ref (shell_client) : NULL;
-    k->use_shell_client = !!shell_client;
 
     k->action = 0;
 
@@ -283,14 +277,18 @@ _keybinding_new (GPasteKeybinding       *binding,
 }
 
 static void
-_keybinding_cleanup_shell_stuff (_Keybinding *k)
+_keybinding_cleanup_shell_stuff (gpointer data,
+                                 gpointer user_data G_GNUC_UNUSED)
 {
+    _Keybinding *k = data;
     k->action = 0;
 }
 
 static void
-_keybinding_free (_Keybinding *k)
+_keybinding_free (gpointer data,
+                  gpointer user_data G_GNUC_UNUSED)
 {
+    _Keybinding *k = data;
     _keybinding_ungrab (k);
     g_signal_handler_disconnect (k->settings, k->c_signals[C_K_REBIND]);
     g_object_unref (k->binding);
@@ -331,12 +329,9 @@ g_paste_keybinder_grab_keybinding_func (gpointer data,
 static void
 g_paste_keybinder_private_grab_all (GPasteKeybinderPrivate *priv)
 {
-    if (!priv->grabbing)
-    {
-        g_slist_foreach (priv->keybindings,
-                         g_paste_keybinder_grab_keybinding_func,
-                         NULL);
-    }
+    g_slist_foreach (priv->keybindings,
+                     g_paste_keybinder_grab_keybinding_func,
+                     NULL);
 }
 
 static void
@@ -356,6 +351,19 @@ retry_grab_all_gnome_shell (gpointer user_data)
     return G_SOURCE_REMOVE;
 }
 
+static void
+release_shell_client (GPasteKeybinderPrivate *priv)
+{
+    g_signal_handler_disconnect (priv->shell_client, priv->c_signals[C_ACCEL]);
+    g_clear_object (&priv->shell_client);
+
+    for (GSList *binding = priv->keybindings; binding; binding = g_slist_next (binding))
+    {
+        _Keybinding *k = binding->data;
+        g_clear_object (&k->shell_client);
+    }
+}
+
 static void
 grab_accelerators_cb (GObject      *source_object,
                       GAsyncResult *res,
@@ -367,11 +375,17 @@ grab_accelerators_cb (GObject      *source_object,
 
     if (error)
     {
-        if (error->code == 19 && priv->retries < 10)
+        if (error->code == G_DBUS_ERROR_UNKNOWN_METHOD && priv->retries < 10)
         {
             ++priv->retries;
             g_source_set_name_by_id (g_timeout_add_seconds (1, retry_grab_all_gnome_shell, priv), "[GPaste] gnome-shell grab");
         }
+        else if (error->code == G_DBUS_ERROR_ACCESS_DENIED)
+        {
+            g_warning ("Couldn't grab keybindings with gnome-shell, falling back to X11 keybinder: %s", error->message);
+            release_shell_client (priv);
+            g_paste_keybinder_private_grab_all (priv);
+        }
         else
         {
             priv->retries = 0;
@@ -439,7 +453,7 @@ g_paste_keybinder_activate_all (GPasteKeybinder *self)
 
     GPasteKeybinderPrivate *priv = g_paste_keybinder_get_instance_private (self);
 
-    if (priv->use_shell_client)
+    if (priv->shell_client)
     {
         g_slist_foreach (priv->keybindings,
                          g_paste_keybinder_activate_keybinding_func,
@@ -605,7 +619,7 @@ on_shell_appeared (GDBusConnection *connection G_GNUC_UNUSED,
 static void
 g_paste_keybinder_private_cleanup_shell_stuff (GPasteKeybinderPrivate *priv)
 {
-    g_slist_foreach (priv->keybindings, (GFunc) _keybinding_cleanup_shell_stuff, NULL);
+    g_slist_foreach (priv->keybindings, _keybinding_cleanup_shell_stuff, NULL);
 }
 
 static void
@@ -632,16 +646,13 @@ g_paste_keybinder_dispose (GObject *object)
     }
 
     if (priv->shell_client)
-    {
-        g_signal_handler_disconnect (priv->shell_client, priv->c_signals[C_ACCEL]);
-        g_clear_object (&priv->shell_client);
-    }
+        release_shell_client (priv);
 
     if (priv->settings)
     {
         g_clear_object (&priv->settings);
         g_paste_keybinder_deactivate_all (self);
-        g_slist_foreach (priv->keybindings, (GFunc) _keybinding_free, NULL);
+        g_slist_foreach (priv->keybindings, _keybinding_free, NULL);
         priv->keybindings = NULL;
     }
 
@@ -710,8 +721,7 @@ g_paste_keybinder_new (GPasteSettings         *settings,
     GPasteKeybinderPrivate *priv = g_paste_keybinder_get_instance_private (self);
 
     priv->settings = g_object_ref (settings);
-    priv->shell_client = (shell_client) ? g_object_ref (shell_client) : NULL;
-    priv->use_shell_client = !!shell_client;
+    priv->shell_client = shell_client = (shell_client) ? g_object_ref (shell_client) : NULL;
     priv->grabbing = FALSE;
     priv->retries = 0;
 
