@@ -201,6 +201,8 @@ G_PASTE_VISIBLE void
 g_paste_util_activate_ui (const gchar *action,
                           GVariant    *arg)
 {
+    g_return_if_fail (g_utf8_validate (action, -1, NULL));
+
     gpointer *data = g_new (gpointer, 2);
     data[0] = g_strdup (action);
     data[1] = arg;
@@ -268,6 +270,73 @@ g_paste_util_activate_ui_sync (const gchar *action,
     return TRUE;
 }
 
+/**
+ * g_paste_util_empty_with_confirmation:
+ * @client: a #GPasteClient instance
+ * @settings: a #GPasteSettings instance
+ * @history: the name of the history to empty
+ *
+ * Empty a history after confirmation.
+ * Confirmation is skipped if GPaste is configured to do so.
+ */
+G_PASTE_VISIBLE void
+g_paste_util_empty_with_confirmation (GPasteClient         *client,
+                                      const GPasteSettings *settings,
+                                      const gchar          *history)
+{
+    g_return_if_fail (_G_PASTE_IS_CLIENT (client));
+    g_return_if_fail (_G_PASTE_IS_SETTINGS (settings));
+    g_return_if_fail (g_utf8_validate (history, -1, NULL));
+
+    if (g_paste_settings_get_empty_history_confirmation (settings))
+        g_paste_util_activate_ui ("empty", g_variant_new_string (history));
+    else
+        g_paste_client_empty_history (client, history, NULL, NULL);
+}
+
+/**
+ * g_paste_util_empty_with_confirmation_sync:
+ * @client: a #GPasteClient instance
+ * @settings: a #GPasteSettings instance
+ * @history: the name of the history to empty
+ * @error: a #GError or %NULL
+ *
+ * Empty a history after confirmation.
+ * Confirmation is skipped if GPaste is configured to do so.
+ *
+ * Returns: whether the action was successful
+ */
+G_PASTE_VISIBLE gboolean
+g_paste_util_empty_with_confirmation_sync (GPasteClient         *client,
+                                           const GPasteSettings *settings,
+                                           const gchar          *history,
+                                           GError              **error)
+{
+    g_return_val_if_fail (_G_PASTE_IS_CLIENT (client), FALSE);
+    g_return_val_if_fail (_G_PASTE_IS_SETTINGS (settings), FALSE);
+    g_return_val_if_fail (g_utf8_validate (history, -1, NULL), FALSE);
+    g_return_val_if_fail (!error || !(*error), FALSE);
+
+    if (g_paste_settings_get_empty_history_confirmation (settings))
+    {
+        return g_paste_util_activate_ui_sync ("empty", g_variant_new_string (history), error);
+    }
+    else
+    {
+        g_autoptr (GError) _error = NULL;
+
+        g_paste_client_empty_history_sync (client, history, &_error);
+
+        if (error)
+        {
+            *error = _error;
+            _error = NULL;
+        }
+
+        return !_error && !(error && *error);
+    }
+}
+
 /**
  * g_paste_util_relace:
  * @text: the initial text
@@ -325,18 +394,21 @@ g_paste_util_compute_checksum (GdkPixbuf *image)
  * g_paste_util_empty_history:
  * @parent_window: (nullable): the parent #GtkWindow
  * @client: a #GPasteClient instance
+ * @settings: a #GPasteSettings instance
  * @history: the name of the history to empty
  *
  * Empty history after prompting user for confirmation
  */
 G_PASTE_VISIBLE void
-g_paste_util_empty_history (GtkWindow    *parent_window,
-                            GPasteClient *client,
-                            const gchar  *history)
+g_paste_util_empty_history (GtkWindow      *parent_window,
+                            GPasteClient   *client,
+                            GPasteSettings *settings,
+                            const gchar    *history)
 {
-    /* Translators: this is the translation for emptying the history */
-    if (g_paste_util_confirm_dialog (parent_window, _("Empty"), _("Do you really want to empty the history?")))
-        g_paste_client_empty_history (client, history, NULL, NULL);
+    if (!g_paste_settings_get_empty_history_confirmation (settings) ||
+        /* Translators: this is the translation for emptying the history */
+        g_paste_util_confirm_dialog (parent_window, _("Empty"), _("Do you really want to empty the history?")))
+            g_paste_client_empty_history (client, history, NULL, NULL);
 }
 
 /**
@@ -377,29 +449,6 @@ g_paste_util_show_win (GApplication *application)
     }
 }
 
-/**
- * g_paste_util_get_at_result:
- * @variant: a #GVariant
- * @len: the length of the resulting array
- *
- * Get the "at" GVariant as an array of guint64
- *
- * Returns: (array length=len): The resulting array
- */
-G_PASTE_VISIBLE guint64 *
-g_paste_util_get_dbus_at_result (GVariant *variant,
-                                 guint64  *len)
-{
-    guint64 _len;
-    const guint64 *r = g_variant_get_fixed_array (variant, &_len, sizeof (guint64));
-    guint64 *ret = g_memdup (r, _len * sizeof (guint64));
-
-    if (len)
-        *len = _len;
-
-    return ret;
-}
-
 /**
  * g_paste_util_get_au_result:
  * @variant: a #GVariant
@@ -415,7 +464,7 @@ g_paste_util_get_dbus_au_result (GVariant *variant,
 {
     guint64 _len;
     const guint32 *r = g_variant_get_fixed_array (variant, &_len, sizeof (guint32));
-    guint32 *ret = g_memdup (r, _len * sizeof (guint32));
+    guint32 *ret = g_memdup2 (r, _len * sizeof (guint32));
 
     if (len)
         *len = _len;
@@ -423,6 +472,49 @@ g_paste_util_get_dbus_au_result (GVariant *variant,
     return ret;
 }
 
+/**
+ * g_paste_util_get_dbus_item_result:
+ * @variant: a #GVariant
+ *
+ * Get the "(ss)" GVariant as an item
+ *
+ * Returns: (transfer full): The item
+ */
+G_PASTE_VISIBLE GPasteClientItem *
+g_paste_util_get_dbus_item_result (GVariant *variant)
+{
+    const gchar *uuid, *value;
+
+    g_variant_get (variant, "(ss)", &uuid, &value);
+
+    return g_paste_client_item_new (uuid, value);
+}
+
+/**
+ * g_paste_util_get_dbus_items_result:
+ * @variant: a #GVariant
+ *
+ * Get the "a(ss)" GVariant as a list of items
+ *
+ * Returns: (element-type GPasteClientItem) (transfer full): The items
+ */
+G_PASTE_VISIBLE GList *
+g_paste_util_get_dbus_items_result (GVariant *variant)
+{
+    GList *items = NULL;
+    GVariantIter iter;
+    g_autoptr (GVariant) v = NULL;
+
+    g_variant_iter_init (&iter, variant);
+    while ((v = g_variant_iter_next_value (&iter)))
+    {
+        items = g_list_append (items, g_paste_util_get_dbus_item_result (v));
+        g_variant_unref (v);
+    }
+
+    return items;
+}
+
 static gchar *
 g_paste_util_get_runtime_dir (const gchar *component)
 {
@@ -525,7 +617,17 @@ g_paste_util_xml_encode (const gchar *text)
 G_PASTE_VISIBLE gchar *
 g_paste_util_get_history_dir_path (void)
 {
-    return g_build_filename (g_get_user_data_dir (), PACKAGE, NULL);
+    const gchar *user_data_dir = g_get_user_data_dir ();
+    gchar *meson_bug_history_path = g_build_filename (user_data_dir, PACKAGE_NAME, NULL);
+
+    // meson wrongfully defined PACKAGE as PACKAGE_NAME.
+    // use it if it exists, but otherwise use the correct path.
+    if (g_file_test (meson_bug_history_path, G_FILE_TEST_IS_DIR))
+        return meson_bug_history_path;
+
+    g_free (meson_bug_history_path);
+
+    return g_build_filename (user_data_dir, PACKAGE, NULL);
 }
 
 /**
